/*
 * This is an example of the Tcl/C++ interface.
 *
 * It's a very tiny demonstration of abstract base classes and virtual
 * functions.
 *
 * We define some trivial "storage" class. We have a couple of classes
 * that can store a value. The only thing they have in common is that
 * the value can be queried as a string.
 *
 * The base class therefore defines the pure virtual function query,
 * which is then overloaded by the storage classes for an integer,
 * a double and a string. Each derived class also defines a "set"
 * function.
 *
 * On the Tcl level, however, we provide a query function for the
 * base class, which on the C++ level automatically calls the
 * correct derived function.
 * The Tcl level instances of derived functions inherit the base
 * class's query function.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#ifdef HAVE_TCLOBJ
#include <tcl.h>
#include <tclobj.h>
#endif

#include "storage.h"

/*
 * ----------------------------------------------------------------------
 * Base class ("storage") definition
 * ----------------------------------------------------------------------
 */

#ifdef HAVE_TCLOBJ

static const TclObj_Comtab sbctab[] = {
  { 1, "query", "", "query value of storage" },
  { 0, NULL, NULL, NULL }
};

TCL_OBJECT_DECL_ABSTRACT_BODY(storage,Tcl_Object,sbctab,NULL);

int
storage::Tcl_Proc (Tcl_Interp *interp,
		   int argc, char *argv[],
		   int command)
{
  switch (command) {
  case 1: /* query */
    return TCL_OBJECT_RETURN_STRING(interp, query());
  }
  TCL_OBJECT_NO_MATCH(interp);
}

#endif

/*
 * ----------------------------------------------------------------------
 * Derived class #1 ("int_storage") definition
 * ----------------------------------------------------------------------
 */

void
int_storage::set (int theint)
{
  value = theint;
}

char *
int_storage::query (void)
{
  static char temp[42];
  sprintf (temp, "%d", value);
  return temp;
}

#ifdef HAVE_TCLOBJ

static const TclObj_Comtab istab[] = {
  {  1, "new", "",     "create empty integer storage" },
  {  1, "set", "int",  "set value of integer storage" },
  {  0, NULL, NULL, NULL }
};

TCL_OBJECT_DECL_BODY(int_storage,storage,istab,NULL);

/*
 * Still, we need an empty Tcl-level constructor
 */

int_storage::int_storage (Tcl_Interp *interp,
			  int argc, char *argv[],
			  int command)
{
  /* this space intentionally left blank */
}

/*
 * Provide set function
 */

int
int_storage::Tcl_Proc (Tcl_Interp *interp,
		       int argc, char *argv[],
		       int command)
{
  switch (command) {
  case 1: /* set int */
    set (TCL_OBJECT_ARG_INT(interp, argv[2]));
    return TCL_OBJECT_RETURN_STRING(interp, query());
  }
  TCL_OBJECT_NO_MATCH(interp);
}

#endif

/*
 * ----------------------------------------------------------------------
 * Derived class #2 ("double_storage") definition
 * ----------------------------------------------------------------------
 */

void
double_storage::set (double thedouble)
{
  value = thedouble;
}

char *
double_storage::query (void)
{
  static char temp[42];
  sprintf (temp, "%f", value);
  return temp;
}

#ifdef HAVE_TCLOBJ

static const TclObj_Comtab dstab[] = {
  {  1, "new", "",        "create empty double storage" },
  {  1, "set", "double",  "set value of double storage" },
  {  0, NULL, NULL, NULL }
};

TCL_OBJECT_DECL_BODY(double_storage,storage,dstab,NULL);

/*
 * Still, we need an empty Tcl-level constructor
 */

double_storage::double_storage (Tcl_Interp *interp,
				int argc, char *argv[],
				int command)
{
  /* this space for rent */
}

/*
 * Provide set function
 */

int
double_storage::Tcl_Proc (Tcl_Interp *interp,
			  int argc, char *argv[],
			  int command)
{
  switch (command) {
  case 1: /* set int */
    set (TCL_OBJECT_ARG_DOUBLE(interp, argv[2]));
    return TCL_OBJECT_RETURN_STRING(interp, query());
  }
  TCL_OBJECT_NO_MATCH(interp);
}

#endif

/*
 * ----------------------------------------------------------------------
 * Derived class #3 ("string_storage") definition
 * ----------------------------------------------------------------------
 */

string_storage::string_storage (void)
{
  string[0] = '\0';
}

void
string_storage::set (const char *thestring)
{
  strcpy (string, thestring);
}

char *
string_storage::query (void)
{
  return string;
}

#ifdef HAVE_TCLOBJ

static const TclObj_Comtab sstab[] = {
  {  1, "new", "",        "create empty string storage" },
  {  1, "set", "string",  "set value of string storage" },
  {  0, NULL, NULL, NULL }
};

TCL_OBJECT_DECL_BODY(string_storage,storage,sstab,NULL);

/*
 * Still, we need an empty Tcl-level constructor
 */

string_storage::string_storage (Tcl_Interp *interp,
				int argc, char *argv[],
				int command)
{
  /* this is a non-smoking function. please kill your cig during execution */
}

/*
 * Provide set function
 */

int
string_storage::Tcl_Proc (Tcl_Interp *interp,
			  int argc, char *argv[],
			  int command)
{
  switch (command) {
  case 1: /* set int */
    set (TCL_OBJECT_ARG_STRING(interp, argv[2]));
    return TCL_OBJECT_RETURN_STRING(interp, query());
  }
  TCL_OBJECT_NO_MATCH(interp);
}

#endif

#ifdef HAVE_TCLOBJ
TCL_OBJECT_BEGIN_PACKAGE(storage);
TCL_OBJECT_PKG_CLASS(int_storage);
TCL_OBJECT_PKG_CLASS(double_storage);
TCL_OBJECT_PKG_CLASS(string_storage);
TCL_OBJECT_END_PACKAGE();
#endif

